@@ -2,7 +2,7 @@
 /*
  * libopenraw - cr2file.cpp
  *
- * Copyright (C) 2006-2018 Hubert Figuière
+ * Copyright (C) 2006-2020 Hubert Figuière
  * Copyright (C) 2008 Novell, Inc.
  *
  * This library is free software: you can redistribute it and/or
@@ -31,7 +31,7 @@
 
 #include "rawdata.hpp"
 #include "rawfile.hpp"
-#include "cfapattern.hpp"
+#include "mosaicinfo.hpp"
 #include "trace.hpp"
 #include "io/memstream.hpp"
 #include "ifdfilecontainer.hpp"
@@ -41,6 +41,7 @@
 #include "jfifcontainer.hpp"
 #include "ljpegdecompressor.hpp"
 #include "rawfile_private.hpp"
+#include "canon.hpp"
 
 using namespace Debug;
 
@@ -58,6 +59,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0xe80,
       { 6264, -582, -724, -8312, 15948, 2504, -1744, 1919, 8664 } },
+    { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_1DMKIIN),
+      0,
+      0xe80,
+      { 6240, -466, -822, -8180, 15825, 2500, -1801, 1938, 8042 } },
     { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_1DMKIII),
       0,
       0xe80,
@@ -178,6 +183,14 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0x3510,
       { 6939, -1016, -866, -4428, 12473, 2177, -1175, 2178, 6162 } },
+    { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_2000D), // Rebel T7
+      0,
+      0,
+      { 8532, -701, -1167, -4095, 11879, 2508, -797, 2424, 7010 } },
+    { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_3000D), // 4000D, Rebel T100
+      0,
+      0,
+      { 6939, -1016, -866, -4428, 12473, 2177, -1175, 2178, 6162 } },
     { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_5D),
       0,
       0xe6c,
@@ -249,6 +262,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 8532, -701, -1167, -4095, 11879, 2508, -797, 2424, 7010 } },
+    { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_EOS_M6),
+      0,
+      0,
+      { 8532, -701, -1167, -4095, 11879, 2508, -797, 2424, 7010 } },
     { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_EOS_M10),
       0,
       0,
@@ -278,6 +295,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 7474, -2301, -567, -4056, 11456, 2975, -222, 716, 4181 } },
+    { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_G16),
+      0,
+      0,
+      { 8020, -2687, -682, -3704, 11879, 2052, -965, 1921, 5556 } },
     // From DNG Converter 7.1-rc
     { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_G1X),
       0,
@@ -331,17 +352,32 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 7968, -2565, -636, -2873, 10697, 2513, 180, 667, 4211 } },
+    { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_S100V),
+      0,
+      0,
+      { 7968, -2565, -636, -2873, 10697, 2513, 180, 667, 4211 } },
     { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_S110),
       0,
       0,
       { 8039, -2643, -654, -3783, 11230, 2930, -206, 690, 4194 } },
+    { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_S120),
+      0,
+      0,
+      { 6961, -1685, -695, -4625, 12945, 1836, -1114, 2152, 5518 } },
     { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_SX50_HS),
       0,
       0,
       { 12432, -4753, -1247, -2110, 10691, 1629, -412, 1623, 4926 } },
+    { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_SX60_HS),
+      0,
+      0,
+      { 13161, -5451, -1344, -1989, 10654, 1531, -47, 1271, 4955 } },
+    { OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_SX1_IS),
+      0,
+      0,
+      { 6578, -259, -502, -5974, 13030, 3309, -308, 1058, 4970 } },
+
     /*
-    { "Canon EOS-1D Mark II N", 0, 0xe80,
-        { 6240,-466,-822,-8180,15825,2500,-1801,1938,8042 } },
     { "Canon EOS-1DS", 0, 0xe20,
         { 4374,3631,-1743,-7520,15212,2472,-2892,3632,8161 } },
     { "Canon EOS-1D", 0, 0xe20,
@@ -392,7 +428,9 @@ const IfdFile::camera_ids_t Cr2File::s_def[] = {
     { "Canon EOS Rebel T6s", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_REBEL_T6S) },
     { "Canon EOS Rebel T6", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_REBEL_T6) },
     { "Canon EOS Rebel T7i", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_REBEL_T7I) },
+    { "Canon EOS Rebel T7", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_REBEL_T7) },
     { "Canon EOS 1000D", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_1000D) },
+    { "Canon EOS 2000D", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_2000D) },
     { "Canon EOS DIGITAL REBEL XS",
       OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_REBEL_XS) },
     { "Canon EOS 1100D", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_1100D) },
@@ -441,6 +479,7 @@ const IfdFile::camera_ids_t Cr2File::s_def[] = {
     { "Canon PowerShot S90", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_S90) },
     { "Canon PowerShot S95", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_S95) },
     { "Canon PowerShot S100", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_S100) },
+    { "Canon PowerShot S100V", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_S100V) },
     { "Canon PowerShot S110", OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_S110) },
     { "Canon PowerShot SX50 HS",
       OR_MAKE_CANON_TYPEID(OR_TYPEID_CANON_SX50_HS) },
@@ -464,15 +503,157 @@ Cr2File::~Cr2File()
 
 IfdDir::Ref Cr2File::_locateCfaIfd()
 {
-    return m_container->setDirectory(3);
+    if (!isCr2()) {
+        return makerNoteIfd();
+    }
+    auto ifd = m_container->setDirectory(3);
+    if (ifd) {
+        ifd->setType(OR_IFD_RAW);
+    }
+    return ifd;
 }
 
 IfdDir::Ref Cr2File::_locateMainIfd()
 {
-    return m_container->setDirectory(0);
+    auto ifd = m_container->setDirectory(0);
+    if (ifd) {
+        ifd->setType(OR_IFD_MAIN);
+    }
+    return ifd;
 }
 
-::or_error Cr2File::_getRawData(RawData &data, uint32_t options)
+::or_error Cr2File::_locateThumbnail(const IfdDir::Ref & dir,
+                                     std::vector<uint32_t> &list)
+{
+    if (isCr2()) {
+        return IfdFile::_locateThumbnail(dir, list);
+    }
+    ::or_error ret = OR_ERROR_NOT_FOUND;
+    ::or_data_type _type = OR_DATA_TYPE_NONE;
+    uint32_t x = dir->getIntegerValue(IFD::EXIF_TAG_IMAGE_WIDTH).value_or(0);
+    uint32_t y = dir->getIntegerValue(IFD::EXIF_TAG_IMAGE_LENGTH).value_or(0);
+    uint32_t offset = 0;
+    uint32_t byte_count = dir->getValue<uint32_t>(IFD::EXIF_TAG_STRIP_BYTE_COUNTS).value_or(0);
+
+    auto result = dir->getValue<uint32_t>(IFD::EXIF_TAG_STRIP_OFFSETS);
+    bool got_it = result.has_value();
+    if (got_it) {
+        offset = result.value();
+    }
+
+    if (x != 0 && y != 0) {
+        // See bug 72270 - some CR2 have 16 bpc RGB thumbnails.
+        // by default it is RGB8. Unless stated otherwise.
+        bool isRGB8 = true;
+        IfdEntry::Ref entry = dir->getEntry(IFD::EXIF_TAG_BITS_PER_SAMPLE);
+        auto result2 = dir->getEntryArrayValue<uint16_t>(*entry);
+        if (result2) {
+          std::vector<uint16_t> arr = result2.value();
+          for(auto bpc : arr) {
+            isRGB8 = (bpc == 8);
+            if (!isRGB8) {
+              LOGDBG1("bpc != 8, not RGB8 %u\n", bpc);
+              break;
+            }
+          }
+        } else {
+          LOGDBG1("Error getting BPS\n");
+        }
+        if (isRGB8) {
+          _type = OR_DATA_TYPE_PIXMAP_8RGB;
+        }
+      }
+
+    if(_type != OR_DATA_TYPE_NONE) {
+      uint32_t dim = std::max(x, y);
+      offset += dir->container().offset();
+      _addThumbnail(dim, ThumbDesc(x, y, _type,
+                                   offset, byte_count));
+      list.push_back(dim);
+      ret = OR_ERROR_NONE;
+    }
+    return ret;
+}
+
+::or_error Cr2File::getRawDataTif(RawData &data, uint32_t options)
+{
+    const IfdDir::Ref &_cfaIfd = cfaIfd();
+    if (!_cfaIfd) {
+        LOGDBG1("cfa IFD not found\n");
+        return OR_ERROR_NOT_FOUND;
+    }
+
+    auto result = _cfaIfd->getValue<uint32_t>(IFD::MNOTE_CANON_RAW_DATA_OFFSET);
+    if (result.empty()) {
+        LOGDBG1("offset not found\n");
+        return OR_ERROR_NOT_FOUND;
+    }
+    uint32_t offset = result.value();
+    uint32_t byte_length = 0;
+
+    result = _cfaIfd->getValue<uint32_t>(IFD::MNOTE_CANON_RAW_DATA_LENGTH);
+    if (!result.empty()) {
+        byte_length = result.value();
+    } else {
+        LOGDBG1("byte len not found\n");
+        auto len = std::min<off_t>(
+            m_container->size(), std::numeric_limits<uint32_t>::max());
+        byte_length = len - offset;
+    }
+
+    std::vector<uint16_t> slices;
+
+    getRawBytes(data, offset, byte_length, 0, 0, slices, options);
+    return OR_ERROR_NONE;
+}
+
+void Cr2File::getRawBytes(RawData &data, uint32_t offset, uint32_t byte_length,
+                          uint16_t x, uint16_t y,
+                          const std::vector<uint16_t>& slices,
+                          uint32_t options)
+{
+    void *p = data.allocData(byte_length);
+    size_t real_size = m_container->fetchData(p, offset, byte_length);
+    if (real_size < byte_length) {
+        LOGWARN("Size mismatch for data: ignoring.\n");
+    }
+    // they are not all RGGB.
+    // but I don't seem to see where this is encoded.
+    //
+    data.setCfaPatternType(OR_CFA_PATTERN_RGGB);
+    data.setDataType(OR_DATA_TYPE_COMPRESSED_RAW);
+    data.setDimensions(x, y);
+
+    LOGDBG1("In size is %dx%d\n", data.width(), data.height());
+    // decompress if we need
+    if ((options & OR_OPTIONS_DONT_DECOMPRESS) == 0) {
+      IO::Stream::Ptr s(new IO::MemStream((const uint8_t*)data.data(),
+                                          data.size()));
+      s->open(); // TODO check success
+      auto jfif = std::make_unique<JfifContainer>(s, 0);
+      LJpegDecompressor decomp(s.get(), jfif.get());
+      // in fact on Canon CR2 files slices either do not exists
+      // or is 3.
+      if (slices.size() > 1) {
+        decomp.setSlices(slices);
+      }
+      RawDataPtr dData = decomp.decompress();
+      if (dData) {
+        LOGDBG1("Out size is %dx%d\n", dData->width(), dData->height());
+        // must re-set the cfaPattern
+        dData->setCfaPatternType(data.mosaicInfo()->patternType());
+        data.swap(*dData);
+        if (!isCr2()) {
+            // In TIF the raw image width needs to be divided by two
+            uint16_t w = data.width();
+            uint16_t h = data.height();
+            data.setDimensions(w / 2, h);
+        }
+      }
+    }
+}
+
+::or_error Cr2File::getRawDataCr2(RawData &data, uint32_t options)
 {
     const IfdDir::Ref &_cfaIfd = cfaIfd();
     if (!_cfaIfd) {
@@ -500,10 +681,10 @@ IfdDir::Ref Cr2File::_locateMainIfd()
     std::vector<uint16_t> slices;
     IfdEntry::Ref e = _cfaIfd->getEntry(IFD::CR2_TAG_SLICE);
     if (e) {
-        auto result2 = e->getArray<uint16_t>();
+        auto result2 = _cfaIfd->getEntryArrayValue<uint16_t>(*e);
         if (result2) {
             slices = result2.value();
-            LOGDBG1("Found slice entry count %ld\n", slices.size());
+            LOGDBG1("Found slice entry count %lu\n", (LSIZE)slices.size());
         }
     }
 
@@ -527,60 +708,47 @@ IfdDir::Ref Cr2File::_locateMainIfd()
     }
     uint16_t y = result2.value();
 
-    void *p = data.allocData(byte_length);
-    size_t real_size = m_container->fetchData(p, offset, byte_length);
-    if (real_size < byte_length) {
-        LOGWARN("Size mismatch for data: ignoring.\n");
-    }
-    // they are not all RGGB.
-    // but I don't seem to see where this is encoded.
-    //
-    data.setCfaPatternType(OR_CFA_PATTERN_RGGB);
-    data.setDataType(OR_DATA_TYPE_COMPRESSED_RAW);
-    data.setDimensions(x, y);
-
-    LOGDBG1("In size is %dx%d\n", data.width(), data.height());
-    // decompress if we need
-    if ((options & OR_OPTIONS_DONT_DECOMPRESS) == 0) {
-      IO::Stream::Ptr s(new IO::MemStream(data.data(), data.size()));
-      s->open(); // TODO check success
-      std::unique_ptr<JfifContainer> jfif(new JfifContainer(s, 0));
-      LJpegDecompressor decomp(s.get(), jfif.get());
-      // in fact on Canon CR2 files slices either do not exists
-      // or is 3.
-      if (slices.size() > 1) {
-        decomp.setSlices(slices);
-      }
-      RawDataPtr dData = decomp.decompress();
-      if (dData) {
-        LOGDBG1("Out size is %dx%d\n", dData->width(), dData->height());
-        // must re-set the cfaPattern
-        dData->setCfaPatternType(data.cfaPattern()->patternType());
-        data.swap(*dData);
-      }
-    }
+    getRawBytes(data, offset, byte_length, x, y, slices, options);
 
     // get the sensor info
     const IfdDir::Ref &_makerNoteIfd = makerNoteIfd();
-    e = _makerNoteIfd->getEntry(IFD::MNOTE_CANON_SENSORINFO);
-    if (e) {
-        auto result3 = e->getArray<uint16_t>();
-        if (result3) {
-            std::vector<uint16_t> sensorInfo = result3.value();
-            if (sensorInfo.size() > 8) {
-                uint32_t w = sensorInfo[7] - sensorInfo[5];
-                uint32_t h = sensorInfo[8] - sensorInfo[6];
-                data.setRoi(sensorInfo[5], sensorInfo[6], w, h);
-            }
-            else {
-                LOGWARN("sensorInfo is too small: %lu - skipping.\n",
-                        sensorInfo.size());
-            }
-        }
+    auto sensorInfo = canon_get_sensorinfo(_makerNoteIfd);
+    if (sensorInfo) {
+        data.setActiveArea((*sensorInfo)[0], (*sensorInfo)[1],
+                           (*sensorInfo)[2], (*sensorInfo)[3]);
     }
 
     return OR_ERROR_NONE;
 }
+
+::or_error Cr2File::_getRawData(RawData &data, uint32_t options)
+{
+    if (!isCr2()) {
+        return getRawDataTif(data, options);
+    }
+    return getRawDataCr2(data, options);
+}
+
+bool Cr2File::vendorCameraIdLocation(Internals::IfdDir::Ref& ifd, uint16_t& index,
+                            const ModelIdMap*& model_map)
+{
+    auto mn = makerNoteIfd();
+    if (mn) {
+        // There is a camera model ID in the MakerNote tag 0x0010.
+        ifd = mn;
+        index = IFD::MNOTE_CANON_MODEL_ID;
+        model_map = &canon_modelid_map;
+        return true;
+    }
+    return false;
+}
+
+bool Cr2File::isCr2()
+{
+    return ((OR_GET_FILE_TYPEID_CAMERA(typeId()) != OR_TYPEID_CANON_1D)
+            && (OR_GET_FILE_TYPEID_CAMERA(typeId()) != OR_TYPEID_CANON_1DS));
+}
+
 }
 }
 
