@@ -2,7 +2,7 @@
 /*
  * libopenraw - ifddir.cpp
  *
- * Copyright (C) 2006-2017 Hubert Figuière
+ * Copyright (C) 2006-2020 Hubert Figuière
  *
  * This library is free software: you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -28,6 +28,8 @@
 #include "ifdfilecontainer.hpp"
 #include "ifddir.hpp"
 #include "makernotedir.hpp"
+#include "metavalue.hpp"
+#include "exif/exif_tags.hpp"
 
 using namespace Debug;
 
@@ -47,8 +49,12 @@ bool IfdDir::isThumbnail() const
     return result && (result.value() == 1);
 }
 
-IfdDir::IfdDir(off_t _offset, IfdFileContainer &_container)
-    : m_offset(_offset), m_container(_container), m_entries()
+IfdDir::IfdDir(off_t _offset, RawContainer& _container, IfdDirType _type, const TagTable& tag_table)
+    : m_type(_type)
+    , m_offset(_offset), m_container(_container), m_entries()
+    , m_tag_table(&tag_table)
+    , m_base_offset(0)
+    , m_endian(_container.endian())
 {
 }
 
@@ -64,14 +70,14 @@ bool IfdDir::load()
     m_entries.clear();
     file->seek(m_offset, SEEK_SET);
 
-    int16_t numEntries = m_container.readInt16(file).value_or(0);
+    int16_t numEntries = m_container.readInt16(file, m_endian).value_or(0);
     LOGDBG1("num entries %d\n", numEntries);
 
     for (int16_t i = 0; i < numEntries; i++) {
         uint32_t data;
-        auto id = m_container.readUInt16(file);
-        auto type = m_container.readInt16(file);
-        auto count = m_container.readInt32(file);
+        auto id = m_container.readUInt16(file, m_endian);
+        auto type = m_container.readInt16(file, m_endian);
+        auto count = m_container.readInt32(file, m_endian);
         size_t sz = file->read(&data, 4);
         if (id.empty() || type.empty() || count.empty() || sz != 4) {
           LOGERR("Failed to read entry %d\n", i);
@@ -80,7 +86,7 @@ bool IfdDir::load()
         uint16_t n_id = id.value();
         IfdEntry::Ref entry =
           std::make_shared<IfdEntry>(n_id, type.value(),
-                                     count.value(), data, m_container);
+                                     count.value(), data, *this);
         m_entries[n_id] = entry;
     }
 
@@ -101,8 +107,8 @@ Option<uint32_t>
 IfdDir::getIntegerValue(uint16_t id)
 {
     IfdEntry::Ref e = getEntry(id);
-    if (e != nullptr) {
-        return Option<uint32_t>(e->getIntegerArrayItem(0));
+    if (e) {
+        return Option<uint32_t>(getEntryIntegerArrayItemValue(*e, 0));
     }
     return Option<uint32_t>();
 }
@@ -114,7 +120,7 @@ off_t IfdDir::nextIFD()
 
     if (m_entries.size() == 0) {
         file->seek(m_offset, SEEK_SET);
-        numEntries = m_container.readInt16(file).value_or(0);
+        numEntries = m_container.readInt16(file, m_endian).value_or(0);
         LOGDBG1("numEntries =%d shifting %d bytes\n", numEntries, (numEntries * 12) + 2);
     } else {
         numEntries = m_entries.size();
@@ -122,7 +128,7 @@ off_t IfdDir::nextIFD()
 
     file->seek(m_offset + (numEntries * 12) + 2, SEEK_SET);
     // XXX how about we check the error. Even though 0 is not valid.
-    return m_container.readInt32(file).value_or(0);
+    return m_container.readInt32(file, m_endian).value_or(0);
 }
 
 /** The SubIFD is locate at offset found in the field
@@ -133,11 +139,11 @@ IfdDir::Ref IfdDir::getSubIFD(uint32_t idx) const
     IfdEntry::Ref e = getEntry(IFD::EXIF_TAG_SUB_IFDS);
 
     if (e != nullptr) {
-        auto result = e->getArray<uint32_t>();
+        auto result = getEntryArrayValue<uint32_t>(*e);
         if (result) {
             std::vector<uint32_t> offsets = result.value();
             if (idx >= offsets.size()) {
-                Ref ref = std::make_shared<IfdDir>(offsets[idx], m_container);
+                Ref ref = std::make_shared<IfdDir>(offsets[idx], m_container, OR_IFD_SUBIFD);
                 ref->load();
                 return ref;
             }
@@ -153,11 +159,11 @@ Option<std::vector<IfdDir::Ref>> IfdDir::getSubIFDs()
     std::vector<IfdDir::Ref> ifds;
     IfdEntry::Ref e = getEntry(IFD::EXIF_TAG_SUB_IFDS);
     if (e != nullptr) {
-        auto result = e->getArray<uint32_t>();
+        auto result = getEntryArrayValue<uint32_t>(*e);
         if (result) {
             std::vector<uint32_t> offsets = result.value();
             for (auto offset_ : offsets) {
-                Ref ifd = std::make_shared<IfdDir>(offset_, m_container);
+                Ref ifd = std::make_shared<IfdDir>(offset_, m_container, OR_IFD_SUBIFD);
                 ifd->load();
                 ifds.push_back(ifd);
             }
@@ -183,12 +189,12 @@ IfdDir::Ref IfdDir::getExifIFD()
     val_offset += m_container.exifOffsetCorrection();
     LOGDBG1("Exif IFD offset = %u\n", val_offset);
 
-    Ref ref = std::make_shared<IfdDir>(val_offset, m_container);
+    Ref ref = std::make_shared<IfdDir>(val_offset, m_container, OR_IFD_EXIF);
     ref->load();
     return ref;
 }
 
-IfdDir::Ref IfdDir::getMakerNoteIfd()
+IfdDir::Ref IfdDir::getMakerNoteIfd(or_rawfile_type file_type)
 {
     uint32_t val_offset = 0;
     IfdEntry::Ref e = getEntry(IFD::EXIF_TAG_MAKER_NOTE);
@@ -201,7 +207,7 @@ IfdDir::Ref IfdDir::getMakerNoteIfd()
     val_offset += m_container.exifOffsetCorrection();
     LOGDBG1("MakerNote IFD offset = %u\n", val_offset);
 
-    auto ref = MakerNoteDir::createMakerNote(val_offset, m_container);
+    auto ref = MakerNoteDir::createMakerNote(val_offset, m_container, file_type);
     if (ref) {
         ref->load();
     }
@@ -209,5 +215,145 @@ IfdDir::Ref IfdDir::getMakerNoteIfd()
     return ref;
 }
 
+const char* IfdDir::getTagName(uint32_t tag) const
+{
+    auto iter = m_tag_table->find(tag);
+    if (iter != m_tag_table->end()) {
+        return iter->second;
+    }
+    return nullptr;
+}
+
+size_t IfdDir::getEntryData(IfdEntry& e, uint8_t* buffer, size_t buffersize) const
+{
+    auto loaded = e.loadDataInto(buffer, buffersize, m_base_offset);
+    return loaded;
+}
+
+uint32_t IfdDir::getEntryIntegerArrayItemValue(IfdEntry& e, int idx) const
+{
+    uint32_t v = 0;
+
+    try {
+        switch (e.type())
+        {
+        case IFD::EXIF_FORMAT_LONG:
+            v = getEntryValue<uint32_t>(e, idx);
+            break;
+        case IFD::EXIF_FORMAT_SHORT:
+            v = getEntryValue<uint16_t>(e, idx);
+            break;
+        case IFD::EXIF_FORMAT_RATIONAL:
+        {
+            IFD::ORRational r = getEntryValue<IFD::ORRational>(e, idx);
+            if (r.denom == 0) {
+                v = 0;
+            } else {
+                v = r.num / r.denom;
+            }
+            break;
+        }
+        default:
+            break;
+        }
+    }
+    catch (const std::exception & ex) {
+        LOGERR("Exception raised %s fetch integer value for %d\n", ex.what(), e.id());
+    }
+
+    return v;
+}
+
+namespace {
+
+template<class T>
+void convert(const IfdDir& dir, Internals::IfdEntry& e, std::vector<MetaValue::value_t>& values)
+{
+    auto result = dir.getEntryArrayValue<T>(e);
+    LOGASSERT(!!result);
+    if (result) {
+        std::vector<T> v = result.value();
+        values.insert(values.end(), v.cbegin(), v.cend());
+    }
+}
+
+// T is the Ifd primitive type. T2 is the target MetaValue type.
+template<class T, class T2>
+void convert(const IfdDir& dir, Internals::IfdEntry& e, std::vector<MetaValue::value_t>& values)
+{
+    auto result = dir.getEntryArrayValue<T>(e);
+    LOGASSERT(!!result);
+    if (result) {
+        std::vector<T> v = result.value();
+        for (const auto & elem : v) {
+            values.push_back(T2(elem));
+        }
+    }
+}
+
+}
+
+MetaValue* IfdDir::makeMetaValue(IfdEntry& entry) const
+{
+    std::vector<MetaValue::value_t> values;
+
+    switch (entry.type()) {
+    case Internals::IFD::EXIF_FORMAT_BYTE:
+    {
+        convert<uint8_t, uint32_t>(*this, entry, values);
+        break;
+    }
+    case Internals::IFD::EXIF_FORMAT_ASCII:
+    {
+        convert<std::string>(*this, entry, values);
+        break;
+    }
+    case Internals::IFD::EXIF_FORMAT_SHORT:
+    {
+        convert<uint16_t, uint32_t>(*this, entry, values);
+        break;
+    }
+    case Internals::IFD::EXIF_FORMAT_LONG:
+    {
+        convert<uint32_t>(*this, entry, values);
+        break;
+    }
+    case Internals::IFD::EXIF_FORMAT_RATIONAL:
+    {
+        convert<Internals::IFD::ORRational>(*this, entry, values);
+        break;
+    }
+    case Internals::IFD::EXIF_FORMAT_SBYTE:
+    {
+        convert<int8_t, int32_t>(*this, entry, values);
+        break;
+    }
+    case Internals::IFD::EXIF_FORMAT_UNDEFINED:
+    {
+        convert<uint8_t>(*this, entry, values);
+        break;
+    }
+    case Internals::IFD::EXIF_FORMAT_SSHORT:
+    {
+        convert<int16_t, int32_t>(*this, entry, values);
+        break;
+    }
+    case Internals::IFD::EXIF_FORMAT_SLONG:
+    {
+        convert<int32_t>(*this, entry, values);
+        break;
+    }
+    case Internals::IFD::EXIF_FORMAT_SRATIONAL:
+    {
+        convert<Internals::IFD::ORSRational>(*this, entry, values);
+        break;
+    }
+    default:
+        LOGDBG1("unhandled type %d\n", type());
+        return nullptr;
+    }
+    return new MetaValue(values);
+}
+
 }
 }
