@@ -1,7 +1,7 @@
 /*
  * libopenraw - ifdfile.cpp
  *
- * Copyright (C) 2006-2017 Hubert Figuière
+ * Copyright (C) 2006-2020 Hubert Figuière
  * Copyright (C) 2008 Novell, Inc.
  *
  * This library is free software: you can redistribute it and/or
@@ -45,6 +45,7 @@
 #include "jfifcontainer.hpp"
 #include "rawfile_private.hpp"
 #include "unpack.hpp"
+#include "xtranspattern.hpp"
 
 namespace OpenRaw {
 
@@ -69,44 +70,104 @@ IfdFile::~IfdFile()
   delete m_container;
 }
 
-// this one seems to be pretty much the same for all the
-// IFD based raw files
-IfdDir::Ref  IfdFile::_locateExifIfd()
+IfdDir::Ref IfdFile::_locateCfaIfd()
 {
-	const IfdDir::Ref & _mainIfd = mainIfd();
-  if (!_mainIfd) {
-    LOGERR("IfdFile::_locateExifIfd() main IFD not found\n");
-    return IfdDir::Ref();
-  }
-  return _mainIfd->getExifIFD();
+  // CFA IFD is the main IFD byt default
+  return mainIfd();
 }
 
-MakerNoteDir::Ref  IfdFile::_locateMakerNoteIfd()
+IfdDir::Ref IfdFile::_locateMainIfd()
 {
-	const IfdDir::Ref & _exifIfd = exifIfd();
-	if(_exifIfd) {
-		// to not have a recursive declaration, getMakerNoteIfd() return an IfdDir.
-		return std::dynamic_pointer_cast<MakerNoteDir>(_exifIfd->getMakerNoteIfd());
-	}
-	return MakerNoteDir::Ref();
+  auto ifd = m_container->setDirectory(0);
+  if (ifd) {
+    ifd->setType(OR_IFD_MAIN);
+  }
+  return ifd;
+}
+
+namespace {
+
+
 }
 
 void IfdFile::_identifyId()
 {
-	const IfdDir::Ref & _mainIfd = mainIfd();
-  if (!_mainIfd) {
-    LOGERR("Main IFD not found to identify the file.\n");
-    return;
-  }
+    // Identify from the vendor internal ID.
+    IfdDir::Ref ifd;
+    uint16_t index = 0;
+    const ModelIdMap* model_map = nullptr;
+    if (vendorCameraIdLocation(ifd, index, model_map) && ifd) {
+        auto id = ifd->getIntegerValue(index);
+        if (id) {
+            auto id_value = id.value();
+            auto type_id = modelid_to_typeid(*model_map, id_value);
+            if (type_id != 0) {
+                _setTypeId(type_id);
+                return;
+            }
+            LOGERR("unknown model ID 0x%x\n", id_value);
+        }
+    }
 
-  auto make = _mainIfd->getValue<std::string>(IFD::EXIF_TAG_MAKE);
-  auto model = _mainIfd->getValue<std::string>(IFD::EXIF_TAG_MODEL);
-  if (make && model) {
-    _setTypeId(_typeIdFromModel(make.value(), model.value()));
-  }
+    // Fallback on using strings.
+    const IfdDir::Ref & _mainIfd = mainIfd();
+    if (!_mainIfd) {
+        LOGERR("Main IFD not found to identify the file.\n");
+        return;
+    }
+
+    auto make = _mainIfd->getValue<std::string>(IFD::EXIF_TAG_MAKE);
+    auto model = _mainIfd->getValue<std::string>(IFD::EXIF_TAG_MODEL);
+    if (!model) {
+        // BlackMagic CinemaDNG doesn't have Make and Model.
+        model = _mainIfd->getValue<std::string>(IFD::DNG_TAG_UNIQUE_CAMERA_MODEL);
+        if (!make) {
+            make = model;
+        }
+    }
+    if (make && model) {
+        _setTypeId(_typeIdFromModel(make.value(), model.value()));
+    }
 }
 
+::or_error IfdFile::_addThumbnailFromStream(uint32_t offset, uint32_t len,
+                                            std::vector<uint32_t>& list)
+{
+  auto err = OR_ERROR_NOT_FOUND;
+  LOGDBG1("fetching JPEG\n");
+  IO::Stream::Ptr s = std::make_shared<IO::StreamClone>(m_io, offset);
+  auto jfif = std::make_unique<JfifContainer>(s, 0);
 
+  uint32_t x, y;
+  x = y = 0;
+  jfif->getDimensions(x, y);
+  LOGDBG1("JPEG dimensions x=%d y=%d\n", x, y);
+
+  uint32_t dim = std::max(x, y);
+  // "Olympus" MakerNote carries a 160 px thubnail we might already have.
+  // We don't check it is the same.
+  if (dim && std::find(list.begin(), list.end(), dim) == list.end()) {
+    _addThumbnail(dim, ThumbDesc(x, y, OR_DATA_TYPE_JPEG, offset, len));
+    list.push_back(dim);
+    err = OR_ERROR_NONE;
+  }
+
+  return err;
+}
+
+::or_error IfdFile::_addThumbnailFromEntry(const IfdEntry::Ref& e, off_t offset,
+                                           std::vector<uint32_t>& list)
+{
+  ::or_error err = OR_ERROR_NOT_FOUND;
+  if (e) {
+    auto val_offset = e->offset();
+
+    val_offset += offset;
+
+    err =_addThumbnailFromStream(val_offset, e->count(), list);
+  }
+  return err;
+}
 
 ::or_error
 IfdFile::_enumThumbnailSizes(std::vector<uint32_t> &list)
@@ -116,7 +177,7 @@ IfdFile::_enumThumbnailSizes(std::vector<uint32_t> &list)
   LOGDBG1("_enumThumbnailSizes()\n");
   std::vector<IfdDir::Ref> & dirs = m_container->directories();
 
-  LOGDBG1("num of dirs %lu\n", dirs.size());
+  LOGDBG1("num of dirs %lu\n", (LSIZE)dirs.size());
   for(auto dir : dirs)
   {
     dir->load();
@@ -156,10 +217,7 @@ IfdFile::_enumThumbnailSizes(std::vector<uint32_t> &list)
 
   auto result = dir->getValue<uint32_t>(IFD::EXIF_TAG_NEW_SUBFILE_TYPE);
   if (result.empty()) {
-    if(!m_cfaIfd) {
-      m_cfaIfd = _locateCfaIfd();
-    }
-    if(m_cfaIfd == dir) {
+    if (cfaIfd() == dir) {
       return OR_ERROR_NOT_FOUND;
     }
     else {
@@ -216,9 +274,8 @@ IfdFile::_enumThumbnailSizes(std::vector<uint32_t> &list)
           _type = OR_DATA_TYPE_JPEG;
           LOGDBG1("looking for JPEG at %u\n", offset);
           if (x == 0 || y == 0) {
-            IO::Stream::Ptr s(std::make_shared<IO::StreamClone>(
-                                m_io, offset));
-            std::unique_ptr<JfifContainer> jfif(new JfifContainer(s, 0));
+            auto s =std::make_shared<IO::StreamClone>(m_io, offset);
+            auto jfif = std::make_unique<JfifContainer>(s, 0);
             if (jfif->getDimensions(x,y)) {
               LOGDBG1("JPEG dimensions x=%u y=%u\n", x, y);
             }
@@ -245,7 +302,7 @@ IfdFile::_enumThumbnailSizes(std::vector<uint32_t> &list)
         // by default it is RGB8. Unless stated otherwise.
         bool isRGB8 = true;
         IfdEntry::Ref entry = dir->getEntry(IFD::EXIF_TAG_BITS_PER_SAMPLE);
-        auto result2 = entry->getArray<uint16_t>();
+        auto result2 = dir->getEntryArrayValue<uint16_t>(*entry);
         if (result2) {
           std::vector<uint16_t> arr = result2.value();
           for(auto bpc : arr) {
@@ -313,8 +370,8 @@ MetaValue *IfdFile::_getMetaValue(int32_t meta_index)
     LOGDBG1("Meta value for %u\n", META_NS_MASKOUT(meta_index));
 
     IfdEntry::Ref e = ifd->getEntry(META_NS_MASKOUT(meta_index));
-    if(e) {
-      val = e->make_meta_value();
+    if (e) {
+      val = ifd->makeMetaValue(*e);
     }
   }
   return val;
@@ -327,54 +384,19 @@ uint32_t IfdFile::_translateCompressionType(IFD::TiffCompress tiff_compression)
 	return (uint32_t)tiff_compression;
 }
 
-
-
-const IfdDir::Ref & IfdFile::cfaIfd()
-{
-	if(!m_cfaIfd) {
-		m_cfaIfd = _locateCfaIfd();
-	}
-	return m_cfaIfd;
-}
-
-
-const IfdDir::Ref & IfdFile::mainIfd()
-{
-	if(!m_mainIfd) {
-		m_mainIfd = _locateMainIfd();
-	}
-	return m_mainIfd;
-}
-
-
-const IfdDir::Ref & IfdFile::exifIfd()
-{
-	if(!m_exifIfd) {
-		m_exifIfd = _locateExifIfd();
-	}
-	return m_exifIfd;
-}
-
-
-const MakerNoteDir::Ref & IfdFile::makerNoteIfd()
-{
-	if(!m_makerNoteIfd) {
-		m_makerNoteIfd = _locateMakerNoteIfd();
-	}
-	return m_makerNoteIfd;
-}
-
-
 namespace {
 
-::or_cfa_pattern
+const MosaicInfo *
 _convertArrayToCfaPattern(const std::vector<uint8_t> &cfaPattern)
 {
   ::or_cfa_pattern cfa_pattern = OR_CFA_PATTERN_NON_RGB22;
   if(cfaPattern.size() != 4) {
-    LOGWARN("Unsupported bayer pattern\n");
-  }
-  else {
+    if (cfaPattern.size() == 36) {
+      // XXX don't assume this is X-Trans
+      return XTransPattern::xtransPattern();
+    }
+    LOGWARN("Unsupported bayer pattern of size %lu\n", (LSIZE)cfaPattern.size());
+  } else {
     LOGDBG2("pattern is = %d, %d, %d, %d\n", cfaPattern[0],
             cfaPattern[1], cfaPattern[2], cfaPattern[3]);
     switch(cfaPattern[0]) {
@@ -393,7 +415,7 @@ _convertArrayToCfaPattern(const std::vector<uint8_t> &cfaPattern)
           cfa_pattern = OR_CFA_PATTERN_GRBG;
         }
         break;
-      case 2:
+      case IFD::CFA_BLUE:
         if ((cfaPattern[2] == IFD::CFA_RED)
             && (cfaPattern[3] == IFD::CFA_GREEN)) {
           cfa_pattern = OR_CFA_PATTERN_GBRG;
@@ -410,65 +432,66 @@ _convertArrayToCfaPattern(const std::vector<uint8_t> &cfaPattern)
       break;
     }
     //
+    return MosaicInfo::twoByTwoPattern(cfa_pattern);
   }
-  return cfa_pattern;
+  return nullptr;
 }
 
-::or_cfa_pattern _convertNewCfaPattern(const IfdEntry::Ref & e)
+const MosaicInfo *_convertNewCfaPattern(const IfdDir::Ref& dir, IfdEntry& e)
 {
-  ::or_cfa_pattern cfa_pattern = OR_CFA_PATTERN_NONE;
-  if(!e || (e->count() < 4)) {
-    return cfa_pattern;
+  if (e.count() < 4) {
+    return nullptr;
   }
 
-  uint16_t hdim = IfdTypeTrait<uint16_t>::get(*e, 0, true);
-  uint16_t vdim = IfdTypeTrait<uint16_t>::get(*e, 1, true);
+  uint16_t hdim = dir->getEntryValue<uint16_t>(e, 0, true);
+  uint16_t vdim = dir->getEntryValue<uint16_t>(e, 1, true);
   if(hdim != 2 && vdim != 2) {
-    cfa_pattern = OR_CFA_PATTERN_NON_RGB22;
-  }
-  else {
+    // cfa_pattern = OR_CFA_PATTERN_NON_RGB22;
+    if (hdim != 6 && vdim != 6) {
+      LOGWARN("CFA pattern dimension %dx%d are incompatible", hdim, vdim);
+      return nullptr;
+    }
+    return XTransPattern::xtransPattern();
+  } else {
     std::vector<uint8_t> cfaPattern;
-    cfaPattern.push_back(IfdTypeTrait<uint8_t>::get(*e, 4, true));
-    cfaPattern.push_back(IfdTypeTrait<uint8_t>::get(*e, 5, true));
-    cfaPattern.push_back(IfdTypeTrait<uint8_t>::get(*e, 6, true));
-    cfaPattern.push_back(IfdTypeTrait<uint8_t>::get(*e, 7, true));
-    cfa_pattern = _convertArrayToCfaPattern(cfaPattern);
+    cfaPattern.push_back(dir->getEntryValue<uint8_t>(e, 4, true));
+    cfaPattern.push_back(dir->getEntryValue<uint8_t>(e, 5, true));
+    cfaPattern.push_back(dir->getEntryValue<uint8_t>(e, 6, true));
+    cfaPattern.push_back(dir->getEntryValue<uint8_t>(e, 7, true));
+    return _convertArrayToCfaPattern(cfaPattern);
   }
-  return cfa_pattern;
 }
 
 
-/** convert the CFA Pattern as stored in the entry */
-::or_cfa_pattern _convertCfaPattern(const IfdEntry::Ref & e)
-{
-  ::or_cfa_pattern cfa_pattern = OR_CFA_PATTERN_NONE;
-
-  auto result = e->getArray<uint8_t>();
-  if (result) {
-    cfa_pattern = _convertArrayToCfaPattern(result.value());
-  }
-
-  return cfa_pattern;
-}
-
-/** get the CFA Pattern out of the directory
- * @param dir the directory
- * @return the cfa_pattern value. %OR_CFA_PATTERN_NONE mean that
- * nothing has been found.
- */
-static ::or_cfa_pattern _getCfaPattern(const IfdDir::Ref & dir)
+/** Extract the MosaicInfo from the entry */
+const MosaicInfo *_convertCfaPattern(const IfdDir::Ref& dir, IfdEntry& e)
 {
   LOGDBG1("%s\n", __FUNCTION__);
-  ::or_cfa_pattern cfa_pattern = OR_CFA_PATTERN_NONE;
+  auto result = dir->getEntryArrayValue<uint8_t>(e);
+  if (result) {
+    return _convertArrayToCfaPattern(result.value());
+  }
+
+  return nullptr;
+}
+
+/** get the mosaic info out of the directory
+ * @param dir the directory
+ * @return the %MosaicInfo* value. %nullptr mean that
+ * nothing has been found.
+ */
+const MosaicInfo *_getMosaicInfo(const IfdDir::Ref & dir)
+{
+  LOGDBG1("%s\n", __FUNCTION__);
+  const MosaicInfo *mosaic_info = nullptr;
   try {
     IfdEntry::Ref e = dir->getEntry(IFD::EXIF_TAG_CFA_PATTERN);
-    if(e) {
-      cfa_pattern = _convertCfaPattern(e);
-    }
-    else {
+    if (e) {
+      mosaic_info = _convertCfaPattern(dir, *e);
+    } else {
       e = dir->getEntry(IFD::EXIF_TAG_NEW_CFA_PATTERN);
-      if(e)  {
-        cfa_pattern = _convertNewCfaPattern(e);
+      if (e)  {
+        mosaic_info = _convertNewCfaPattern(dir, *e);
       }
     }
   }
@@ -476,7 +499,7 @@ static ::or_cfa_pattern _getCfaPattern(const IfdDir::Ref & dir)
   {
     LOGERR("Exception in _getCfaPattern().\n");
   }
-  return cfa_pattern;
+  return mosaic_info;
 }
 
 } // end anon namespace
@@ -532,7 +555,7 @@ static ::or_cfa_pattern _getCfaPattern(const IfdDir::Ref & dir)
       LOGDBG1("byte len not found\n");
       return OR_ERROR_NOT_FOUND;
     }
-    auto result3 = e->getArray<uint32_t>();
+    auto result3 = dir->getEntryArrayValue<uint32_t>(*e);
     if (result3) {
       std::vector<uint32_t> counts = result3.value();
       LOGDBG1("counting tiles\n");
@@ -547,7 +570,7 @@ static ::or_cfa_pattern _getCfaPattern(const IfdDir::Ref & dir)
       LOGDBG1("tile offsets empty\n");
       return OR_ERROR_NOT_FOUND;
     }
-    auto result3 = e->getArray<uint32_t>();
+    auto result3 = dir->getEntryArrayValue<uint32_t>(*e);
     if (!result3) {
       LOGDBG1("tile offsets not found\n");
       return OR_ERROR_NOT_FOUND;
@@ -559,7 +582,7 @@ static ::or_cfa_pattern _getCfaPattern(const IfdDir::Ref & dir)
       LOGDBG1("tile byte counts not found\n");
       return OR_ERROR_NOT_FOUND;
     }
-    result3 = e->getArray<uint32_t>();
+    result3 = dir->getEntryArrayValue<uint32_t>(*e);
     if (result3) {
       std::vector<uint32_t> counts = result3.value();
       LOGDBG1("counting tiles\n");
@@ -610,16 +633,16 @@ static ::or_cfa_pattern _getCfaPattern(const IfdDir::Ref & dir)
   LOGDBG1("RAW Compression is %u\n", compression);
   LOGDBG1("bpc is %u\n", bpc);
 
-  ::or_cfa_pattern cfa_pattern = _getCfaPattern(dir);
-  if(cfa_pattern == OR_CFA_PATTERN_NONE) {
+  const MosaicInfo *mosaic_info = _getMosaicInfo(dir);
+  if (!mosaic_info) {
     // some file have it in the exif IFD instead.
-    if(!m_exifIfd) {
-      m_exifIfd = _locateExifIfd();
-    }
-    cfa_pattern = _getCfaPattern(m_exifIfd);
+    mosaic_info = _getMosaicInfo(exifIfd());
   }
 
 
+  // This is the actual bit per component since we might readjust
+  // for padding.
+  auto actual_bpc = bpc;
   if((bpc == 12 || bpc == 14) && (compression == IFD::COMPRESS_NONE)
      && (byte_length == (x * y * 2))) {
     // We turn this to a 16-bits per sample. MSB are 0
@@ -642,14 +665,14 @@ static ::or_cfa_pattern _getCfaPattern(const IfdDir::Ref & dir)
     LOGERR("Unsupported bpc %u\n", bpc);
     return OR_ERROR_INVALID_FORMAT;
   }
-  data.setCfaPatternType(cfa_pattern);
+  data.setMosaicInfo(mosaic_info);
   data.setDataType(data_type);
   data.setBpc(bpc);
   data.setCompression(data_type == OR_DATA_TYPE_COMPRESSED_RAW
                       ? compression : 1);
   data.setPhotometricInterpretation((ExifPhotometricInterpretation)photo_int);
-  if((data_type == OR_DATA_TYPE_RAW) && (data.whiteLevel() == 0)) {
-    data.setWhiteLevel((1 << bpc) - 1);
+  if(data.whiteLevel() == 0) {
+    data.setWhiteLevel((1 << actual_bpc) - 1);
   }
   data.setDimensions(x, y);
 
@@ -666,11 +689,11 @@ IfdFile::_unpackData(uint16_t bpc, uint32_t compression, RawData & data,
   uint32_t current_offset = offset;
   Unpack unpack(x, compression);
   const size_t blocksize = (bpc == 8 ? x : unpack.block_size());
-  LOGDBG1("Block size = %lu\n", blocksize);
+  LOGDBG1("Block size = %lu\n", (LSIZE)blocksize);
   LOGDBG1("dimensions (x, y) %u, %u\n", x, y);
-  std::unique_ptr<uint8_t[]> block(new uint8_t[blocksize]);
+  auto block = std::make_unique<uint8_t[]>(blocksize);
   size_t outsize = x * y * 2;
-  uint8_t * outdata = (uint8_t*)data.allocData(outsize);
+  uint16_t* outdata = (uint16_t*)data.allocData(outsize);
   size_t got;
   LOGDBG1("offset of RAW data = %u\n", current_offset);
   do {
@@ -679,23 +702,20 @@ IfdFile::_unpackData(uint16_t bpc, uint32_t compression, RawData & data,
     fetched += got;
     offset += got;
     current_offset += got;
-    if(got) {
-      if(bpc == 12) {
+    if (got) {
+      if (bpc == 12) {
         size_t out;
         ret = unpack.unpack_be12to16(outdata, outsize,
                                      block.get(),
                                      got, out);
-        outdata += out;
+        outdata += out / 2;
         outsize -= out;
-        if(ret != OR_ERROR_NONE) {
+        if (ret != OR_ERROR_NONE) {
           break;
         }
-      }
-      else {
-        // outdata point to uint16_t
-        std::copy(block.get(), block.get()+got,
-                  (uint16_t*)outdata);
-        outdata += (got << 1);
+      } else {
+        std::copy(block.get(), block.get() + got, outdata);
+        outdata += got;
       }
     }
   } while((got != 0) && (fetched < byte_length));
@@ -711,8 +731,8 @@ IfdFile::_unpackData(uint16_t bpc, uint32_t compression, RawData & data,
   mode:c++
   c-file-style:"stroustrup"
   c-file-offsets:((innamespace . 0))
-  tab-width:2
-  c-basic-offset:2
+  tab-width:4
+  c-basic-offset:4
   indent-tabs-mode:nil
   fill-column:80
   End:
