@@ -1,7 +1,7 @@
 /*
  * libopenraw - memstream.cpp
  *
- * Copyright (C) 2007-2017 Hubert Figuière
+ * Copyright (C) 2007-2018 Hubert Figuière
  *
  * This library is free software: you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -33,7 +33,7 @@ using namespace Debug;
 namespace OpenRaw {
 namespace IO {
 
-MemStream::MemStream(void *ptr, size_t s)
+MemStream::MemStream(const uint8_t* ptr, size_t s)
   : Stream(""),
     m_ptr(ptr),
     m_size(s),
@@ -43,14 +43,14 @@ MemStream::MemStream(void *ptr, size_t s)
 
 or_error MemStream::open()
 {
-  m_current = (unsigned char *)m_ptr;
+  m_current = m_ptr;
   return OR_ERROR_NONE;
 }
 
 
 int MemStream::close()
 {
-  m_current = NULL;
+  m_current = nullptr;
   return 0;
 }
 
@@ -63,19 +63,19 @@ int MemStream::seek(off_t offset, int whence)
     // TODO set error code
     return -1;
   }
-  switch(whence)
+  switch (whence)
   {
   case SEEK_SET:
-    m_current = (unsigned char*)m_ptr + offset;
+    m_current = m_ptr + offset;
     newpos = offset;
     break;
   case SEEK_END:
-    m_current = (unsigned char*)m_ptr + m_size + offset;
+    m_current = m_ptr + m_size + offset;
     newpos = m_size + offset;
     break;
   case SEEK_CUR:
     m_current += offset;
-    newpos = (m_current - (unsigned char*)m_ptr);
+    newpos = (m_current - m_ptr);
     break;
   default:
     return -1;
@@ -87,13 +87,13 @@ int MemStream::seek(off_t offset, int whence)
 
 int MemStream::read(void *buf, size_t count)
 {
-  if((m_current == nullptr) || (m_ptr == nullptr)) {
+  if ((m_current == nullptr) || (m_ptr == nullptr)) {
     LOGDBG1("MemStream::failed\n");
     return -1;
   }
 
-  unsigned char * end = (unsigned char*)m_ptr + m_size;
-  if((off_t)count > (end - m_current)) {
+  const uint8_t* end = m_ptr + m_size;
+  if ((off_t)count > (end - m_current)) {
     count = end - m_current;
     // TODO set EOF
   }
