@@ -2,7 +2,7 @@
 /*
  * libopenraw - neffile.cpp
  *
- * Copyright (C) 2006-2017 Hubert Figuière
+ * Copyright (C) 2006-2020 Hubert Figuière
  * Copyright (C) 2008 Novell, Inc.
  *
  * This library is free software: you can redistribute it and/or
@@ -32,11 +32,12 @@
 #include <libopenraw/consts.h>
 #include <libopenraw/debug.h>
 
-#include "cfapattern.hpp"
+#include "io/streamclone.hpp"
+
+#include "mosaicinfo.hpp"
 #include "rawdata.hpp"
 #include "rawfile.hpp"
 
-
 #include "trace.hpp"
 #include "ifd.hpp"
 #include "ifdfilecontainer.hpp"
@@ -46,6 +47,7 @@
 #include "nefcfaiterator.hpp"
 #include "neffile.hpp"
 #include "rawcontainer.hpp"
+#include "jfifcontainer.hpp"
 #include "rawfile_private.hpp"
 
 using namespace Debug;
@@ -79,10 +81,18 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 5710,-901,-615,-8594,16617,2024,-2975,4120,6830 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D2HS),
+      0,
+      0,
+      { 5710, -901, -615, -8594, 16617, 2024, -2975, 4120, 6830 } },
     { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D2X),
       0,
       0,
       { 10231,-2769,-1255,-8301,15900,2552,-797,680,7148 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D2XS),
+      0,
+      0,
+      { 10231, -2769, -1255, -8301, 15900, 2552, -797, 680, 7148 } },
 
     { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3),
       0,
@@ -120,6 +130,14 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 6988, -1384, -714, -5631, 13410, 2447, -1485, 2204, 7318 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3400),
+      0,
+      0,
+      { 6988, -1384, -714, -5631, 13410, 2447, -1485, 2204, 7318 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3500),
+      0,
+      0,
+      { 8821, -2938, -785, -4178, 12142, 2287, -824, 1651, 6860 } },
 
     { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D4),
       0,
@@ -176,6 +194,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 8821, -2938, -785, -4178, 12142, 2287, -824, 1651, 6860 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D6),
+      0,
+      0,
+      { 9028, -3423, -1035, -6321, 14265, 2217, -1013, 1683, 6928 } },
     { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D60),
       0,
       0,
@@ -219,6 +241,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 9020, -2890, -715, -4535, 12436, 2348, -934, 1919, 7086 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D780),
+      0,
+      0,
+      { 9943, -3269, -839, -5323, 13269, 2259, -1198, 2083, 7557 } },
 
     { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D80),
       0,
@@ -244,6 +270,31 @@ static const BuiltinColourMatrix s_matrices[] = {
       0xf00,
       { 7309,-1403,-519,-8474,16008,2622,-2434,2826,8064 } },
 
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z6),
+      0,
+      0,
+      { 8210, -2534, -683, -5355, 13338, 2212, -1143, 1929, 6464 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z6_2),
+      0,
+      0,
+      { 9943, -3269, -839, -5323, 13269, 2259, -1198, 2083, 7557 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z7),
+      0,
+      0,
+      { 10405, -3755, -1270, -5461, 13787, 1793, -1040, 2015, 6785 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z7_2),
+      0,
+      0,
+      { 13705, -6004, -1400, -5464, 13568, 2062, -940, 1706, 7618 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z50),
+      0,
+      0,
+      { 11640, -4829, -1079, -5107, 13006, 2325, -972, 1711, 7380 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z5),
+      0,
+      0,
+      { 8695, -2558, -648, -5015, 12711, 2575, -1279, 2215, 7514 } },
+
     { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_DF),
       0,
       0,
@@ -259,11 +310,19 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 7842,-2320,-992,-8154,15718,2599,-1098,1342,7560 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_E8800),
+      0,
+      0,
+      { 7971, -2314, -913, -8451, 15762, 2894, -1442, 1520, 7610 } },
     // From DNG Converter 10.3
     { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_B700),
       200,
       0,
       { 14387, -6014, -1299, -1357, 9975, 1616, 467, 1047, 4744 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P1000),
+      0,
+      0,
+      { 14294, -6116, -1333, -1628, 10219, 1637, -14, 1158, 5022 } },
     { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P330),
       200,
       0,
@@ -272,6 +331,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       200,
       0,
       { 10321, -3920, -931, -2750, 11146, 1824, -442, 1545, 5539 } },
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P950),
+      0,
+      0,
+      { 13307, -5641, -1290, -2048, 10581, 1689, -64, 1222, 5176 } },
     { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P6000),
       0,
       0,
@@ -343,6 +406,11 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       { 8198,-2239,-724,-4871,12389,2798,-1043,2050,7181 } },
 
+    { OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_A1000),
+      0,
+      0,
+      { 10601, -3487, -1127, -2931, 11443, 1676, -587, 1740, 5278 } },
+
     { 0, 0, 0, { 0, 0, 0, 0, 0, 0, 0, 0, 0 } }
 };
 
@@ -351,8 +419,10 @@ const IfdFile::camera_ids_t NefFile::s_def[] = {
     { "NIKON D100 ", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D100) },
     { "NIKON D1X", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D1X) },
     { "NIKON D200", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D200) },
-    { "NIKON D2H", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D2H ) },
-    { "NIKON D2X", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D2X ) },
+    { "NIKON D2H", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D2H) },
+    { "NIKON D2Hs", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D2HS) },
+    { "NIKON D2X", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D2X) },
+    { "NIKON D2Xs", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D2XS ) },
     { "NIKON D3", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3) },
     { "NIKON D3S", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3S) },
     { "NIKON D3X", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3X) },
@@ -363,6 +433,7 @@ const IfdFile::camera_ids_t NefFile::s_def[] = {
     { "NIKON D3200", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3200) },
     { "NIKON D3300", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3300) },
     { "NIKON D3400", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3400) },
+    { "NIKON D3500", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D3500) },
     { "NIKON D4", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D4) },
     { "NIKON D4S", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D4S) },
     { "NIKON D40", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D40) },
@@ -376,6 +447,7 @@ const IfdFile::camera_ids_t NefFile::s_def[] = {
     { "NIKON D5300", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D5300) },
     { "NIKON D5500", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D5500) },
     { "NIKON D5600", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D5600) },
+    { "NIKON D6", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D6) },
     { "NIKON D60",   OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D60) },
     { "NIKON D600", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D600) },
     { "NIKON D610", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D610) },
@@ -386,23 +458,34 @@ const IfdFile::camera_ids_t NefFile::s_def[] = {
     { "NIKON D7100", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D7100) },
     { "NIKON D7200", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D7200) },
     { "NIKON D750", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D750) },
+    { "NIKON D780", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D780) },
     { "NIKON D80", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D80) },
     { "NIKON D800", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D800) },
     { "NIKON D800E", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D800E) },
     { "NIKON D810", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D810) },
     { "NIKON D90", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_D90) },
     { "NIKON Df", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_DF) },
+    { "NIKON Z 6", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z6) },
+    { "NIKON Z 6_2", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z6_2) },
+    { "NIKON Z 7", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z7) },
+    { "NIKON Z 7_2", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z7_2) },
+    { "NIKON Z 50", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z50) },
+    { "NIKON Z 5", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_Z5) },
     { "E5400", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_E5400) },
     { "E5700", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_E5700) },
     { "E8400", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_E8400) },
+    { "E8800", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_E8800) },
     { "COOLPIX B700", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_B700) },
     { "COOLPIX P330", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P330) },
     { "COOLPIX P340", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P340) },
+    { "COOLPIX P950", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P950) },
+    { "COOLPIX P1000", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P1000) },
     { "COOLPIX P6000", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P6000) },
     { "COOLPIX P7000", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P7000) },
     { "COOLPIX P7100", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P7100) },
     { "COOLPIX P7700", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_P7700) },
     { "COOLPIX A", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_A) },
+    { "COOLPIX A1000", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_COOLPIX_A1000) },
     { "NIKON 1 J1", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_1_J1) },
     { "NIKON 1 J2", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_1_J2) },
     { "NIKON 1 J3", OR_MAKE_NIKON_TYPEID(OR_TYPEID_NIKON_1_J3) },
@@ -455,7 +538,7 @@ bool NefFile::isCompressed(RawContainer & container, uint32_t offset)
 
 bool NefFile::isNrw()
 {
-    MakerNoteDir::Ref _makerNoteIfd = makerNoteIfd();
+    IfdDir::Ref _makerNoteIfd = makerNoteIfd();
     if(!_makerNoteIfd) {
         LOGERR("makernote not found\n");
         return false;
@@ -478,6 +561,28 @@ NefFile::_unpackData(uint16_t bpc, uint32_t compression, RawData & data,
     return TiffEpFile::_unpackData(bpc, compression, data, x, y, offset, byte_length);
 }
 
+::or_error NefFile::_enumThumbnailSizes(std::vector<uint32_t> &list)
+{
+    or_error err = this->TiffEpFile::_enumThumbnailSizes(list);
+    auto makernote = makerNoteIfd();
+    if (makernote) {
+        auto ifd = makernote->getIfdInEntry(MNOTE_NIKON_PREVIEW_IFD);
+        if (ifd) {
+            auto start = ifd->getValue<uint32_t>(MNOTE_NIKON_PREVIEWIFD_START).value_or(0);
+            auto len = ifd->getValue<uint32_t>(MNOTE_NIKON_PREVIEWIFD_LENGTH).value_or(0);
+
+            // if either value is zero we consider it is wrong.
+            if (start != 0 && len != 0) {
+                start += makernote->getMnoteOffset();
+                err = _addThumbnailFromStream(start, len, list);
+            } else {
+                err = OR_ERROR_NOT_FOUND;
+            }
+        }
+    }
+
+    return err;
+}
 
 uint32_t NefFile::_translateCompressionType(IFD::TiffCompress tiffCompression)
 {
@@ -500,14 +605,14 @@ uint32_t NefFile::_translateCompressionType(IFD::TiffCompress tiffCompression)
     }
     const uint32_t rows = data.height();
     const uint32_t raw_columns = data.width();
-    
+
     //FIXME: not always true
     const uint32_t columns = raw_columns - 1;
-    
+
     NefDiffIterator
         diffs(c.huffman, static_cast<uint8_t*>(data.data()), data.size());
-    NefCfaIterator iter(diffs, rows, raw_columns, c.vpred);
-    
+    NefCfaIterator iter(diffs, raw_columns, c.vpred);
+
     RawData newData;
     uint16_t *p = (uint16_t *) newData.allocData(rows * columns * 2);
     newData.setDimensions(columns, rows);
@@ -515,7 +620,7 @@ uint32_t NefFile::_translateCompressionType(IFD::TiffCompress tiffCompression)
     uint16_t bpc = data.bpc();
     newData.setBpc(bpc);
     newData.setWhiteLevel((1 << bpc) - 1);
-    newData.setCfaPatternType(data.cfaPattern()->patternType());
+    newData.setCfaPatternType(data.mosaicInfo()->patternType());
 
     for (unsigned int i = 0; i < rows; i++) {
         for (unsigned int j = 0; j < raw_columns; j++) {
@@ -551,7 +656,7 @@ uint32_t NefFile::_translateCompressionType(IFD::TiffCompress tiffCompression)
 bool
 NefFile::_getCompressionCurve(RawData & data,  NefFile::NEFCompressionInfo& c)
 {
-    MakerNoteDir::Ref _makerNoteIfd = makerNoteIfd();
+    MakerNoteDir::Ref _makerNoteIfd = std::dynamic_pointer_cast<MakerNoteDir>(makerNoteIfd());
     if(!_makerNoteIfd) {
         LOGERR("makernote not found\n");
         return false;
@@ -562,6 +667,8 @@ NefFile::_getCompressionCurve(RawData & data,  NefFile::NEFCompressionInfo& c)
         return false;
     }
 
+    auto endian = m_container->endian();
+
     size_t pos = _makerNoteIfd->getMnoteOffset() + curveEntry->offset();
 
     auto file = m_container->file();
@@ -591,7 +698,7 @@ NefFile::_getCompressionCurve(RawData & data,  NefFile::NEFCompressionInfo& c)
 
     for (int i = 0; i < 2; ++i) {
         for (int j = 0; j < 2; ++j) {
-            auto result16 = m_container->readInt16(file);
+            auto result16 = m_container->readInt16(file, endian);
             if(result16.empty()) {
                 LOGERR("Failed to read vpred (%d,%d)\n", i, j);
                 return false;
@@ -635,8 +742,8 @@ NefFile::_getCompressionCurve(RawData & data,  NefFile::NEFCompressionInfo& c)
 
     // number of elements in the curve
     size_t nelems;
-    nelems = m_container->readInt16(file).value_or(0);
-    LOGDBG1("Num elems %ld\n", nelems);
+    nelems = m_container->readInt16(file, endian).value_or(0);
+    LOGDBG1("Num elems %lu\n", (LSIZE)nelems);
 
     uint32_t ceiling = 1 << bpc & 0x7fff;
     uint32_t step = 0;
@@ -647,7 +754,7 @@ NefFile::_getCompressionCurve(RawData & data,  NefFile::NEFCompressionInfo& c)
 
     if (header0 == 0x44 && header1 == 0x20 && step > 0) {
         for (size_t i = 0; i < nelems; ++i) {
-            auto result16 = m_container->readInt16(file);
+            auto result16 = m_container->readInt16(file, endian);
             if (result16.empty()) {
                 LOGERR("NEF: short read\n");
                 return false;
@@ -663,7 +770,7 @@ NefFile::_getCompressionCurve(RawData & data,  NefFile::NEFCompressionInfo& c)
     } else if (header0 != 0x46 && nelems <= 0x4001) {
         size_t num_read = m_container->readUInt16Array(file, c.curve, nelems);
         if (num_read < nelems) {
-            LOGERR("NEF: short read of %ld elements instead of %ld\n", num_read, nelems);
+            LOGERR("NEF: short read of %lu elements instead of %lu\n", (LSIZE)num_read, (LSIZE)nelems);
             return false;
         }
         ceiling = nelems;
