@@ -1,7 +1,7 @@
 /*
  * libopenraw - orffile.cpp
  *
- * Copyright (C) 2006-2017 Hubert Figuière
+ * Copyright (C) 2006-2020 Hubert Figuière
  *
  * This library is free software: you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -26,7 +26,7 @@
 #include <libopenraw/cameraids.h>
 #include <libopenraw/consts.h>
 
-#include "cfapattern.hpp"
+#include "mosaicinfo.hpp"
 #include "rawdata.hpp"
 #include "rawfile.hpp"
 
@@ -102,18 +102,22 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 12078,-4836,-1069,-6671,14306,2578,-786,939,7418 } },
-    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP500),
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP500UZ),
       0,
       0xfff,
       { 9493,-3415,-666,-5211,12334,3260,-1548,2262,6482 } },
-    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP510),
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP510UZ),
       0,
       0xffe,
       { 10593,-3607,-1010,-5881,13127,3084,-1200,1805,6721 } },
-    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP550),
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP550UZ),
       0,
       0xffe,
       { 11597,-4006,-1049,-5432,12799,2957,-1029,1750,6516 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP565UZ),
+      0,
+      0xffe,
+      { 11856, -4470, -1159, -4814, 12368, 2756, -994, 1780, 5589 } },
     { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EP1),
       0,
       0xffd,
@@ -158,6 +162,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0xfcb,
       { 8380, -2630, -639, -2887, 10725, 2496, -627, 1427, 5438 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EPL10),
+      0,
+      0xfcb,
+      { 9197, -3190, -659, -2606, 10830, 2039, -458, 1250, 5458 } },
     { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EPM1),
       0,
       0,
@@ -186,10 +194,22 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 9422, -3258, -711, -2655, 10898, 2015, -512, 1354, 5512 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM5III),
+      0,
+      0,
+      { 11896, -5110, -1076, -3181, 11378, 2048, -519, 1224, 5166 } },
     { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM1),
       0,
       0,
       { 7687, -1984, -606, -4327, 11928, 2721, -1381, 2339, 6452 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM1II),
+      0,
+      0,
+      { 9383, -3170, -763, -2457, 10702, 2020, -384, 1236, 5552 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM1III),
+      0,
+      0,
+      { 11896, -5110, -1076, -3181, 11378, 2048, -519, 1224, 5166 } },
     { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM10),
       0,
       0,
@@ -198,10 +218,26 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 8380, -2630, -639, -2887, 10725, 2496, -627, 1427, 5438 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM10III),
+      0,
+      0,
+      { 8380, -2630, -639, -2887, 10725, 2496, -627, 1427, 5438 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM10IV),
+      0,
+      0,
+      { 9476, -3182, -765, -2613, 10958, 1893, -449, 1315, 5268 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM1X),
+      0,
+      0,
+      { 11896, -5110, -1076, -3181, 11378, 2048, -519, 1224, 5166 } },
     { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_STYLUS1),
       0,
       0,
       { 8360, -2420, -880, -3928, 12353, 1739, -1381, 2416, 5173 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_STYLUS1_1S),
+      0,
+      0,
+      { 8360, -2420, -880, -3928, 12353, 1739, -1381, 2416, 5173 } },
     { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_PEN_F),
       0,
       0,
@@ -214,6 +250,14 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 11426, -4159, -1126, -2066, 10678, 1593, -120, 1327, 4998 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_TG5),
+      0,
+      0,
+      { 10899, -3833, -1082, -2112, 10736, 1575, -267, 1452, 5269 } },
+    { OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_TG6),
+      0,
+      0,
+      { 10899, -3833, -1082, -2112, 10736, 1575, -267, 1452, 5269 } },
 
     { 0, 0, 0, { 0, 0, 0, 0, 0, 0, 0, 0, 0 } }
 
@@ -232,9 +276,10 @@ const struct IfdFile::camera_ids_t OrfFile::s_def[] = {
     { "E-510           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_E510) },
     { "E-620           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_E620) },
     { "SP350"           , OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP350) },
-    { "SP500UZ"         , OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP500) },
-    { "SP510UZ"         , OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP510) },
-    { "SP550UZ                ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP550) },
+    { "SP500UZ"         , OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP500UZ) },
+    { "SP510UZ"         , OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP510UZ) },
+    { "SP550UZ                ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP550UZ) },
+    { "SP565UZ                ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SP565UZ) },
     { "E-P1            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EP1) },
     { "E-P2            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EP2) },
     { "E-P3            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EP3) },
@@ -246,6 +291,7 @@ const struct IfdFile::camera_ids_t OrfFile::s_def[] = {
     { "E-PL7           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EPL7) },
     { "E-PL8           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EPL8) },
     { "E-PL9           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EPL9) },
+    { "E-PL10          ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EPL10) },
     { "E-PM1           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EPM1) },
     { "E-PM2           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EPM2) },
     { "XZ-1            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_XZ1) },
@@ -253,14 +299,22 @@ const struct IfdFile::camera_ids_t OrfFile::s_def[] = {
     { "XZ-2            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_XZ2) },
     { "E-M5            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM5) },
     { "E-M5MarkII      ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM5II) },
+    { "E-M5MarkIII     ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM5III) },
     { "E-M1            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM1) },
     { "E-M1MarkII      ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM1II) },
+    { "E-M1MarkIII     ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM1III) },
+    { "E-M1X           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM1X) },
     { "E-M10           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM10) },
     { "E-M10MarkII     ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM10II) },
+    { "E-M10 Mark III  ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM10III) },
+    { "E-M10MarkIV     ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_EM10IV) },
     { "STYLUS1         ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_STYLUS1) },
+    { "STYLUS1,1s      ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_STYLUS1_1S) },
     { "PEN-F           ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_PEN_F) },
     { "SH-2            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_SH2) },
     { "TG-4            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_TG4) },
+    { "TG-5            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_TG5) },
+    { "TG-6            ", OR_MAKE_OLYMPUS_TYPEID(OR_TYPEID_OLYMPUS_TG6) },
     { 0, 0 }
 };
 
@@ -281,60 +335,47 @@ OrfFile::~OrfFile()
 {
 }
 
-IfdDir::Ref  OrfFile::_locateCfaIfd()
-{
-    // in ORF the CFA IFD is the main IFD
-    return mainIfd();
-}
-
-
-IfdDir::Ref  OrfFile::_locateMainIfd()
-{
-    return m_container->setDirectory(0);
-}
-
 ::or_error OrfFile::_enumThumbnailSizes(std::vector<uint32_t> &list)
 {
     auto err = OR_ERROR_NOT_FOUND;
 
     err = IfdFile::_enumThumbnailSizes(list);
+    LOGDBG1("got %lu thumbs\n", (LSIZE)list.size());
 
     auto exif = exifIfd();
     if (!exif) {
+        LOGDBG1("can't get exif\n");
         return err;
     }
 
-    auto ifd = exif->getMakerNoteIfd();
-    if (!ifd) {
-        return err;
-    }
-    auto makerNote = std::dynamic_pointer_cast<MakerNoteDir>(ifd);
+    auto makerNote = std::dynamic_pointer_cast<MakerNoteDir>(exif->getMakerNoteIfd(type()));
     if (!makerNote) {
+        LOGDBG1("can't get makernote\n");
         return err;
     }
 
-    auto e = makerNote->getEntry(0x100);
-    if (e) {
-        auto val_offset = e->offset();
+    auto e = makerNote->getEntry(ORF_TAG_THUMBNAIL_IMAGE);
+    _addThumbnailFromEntry(e, makerNote->getMnoteOffset(), list);
 
-        val_offset += makerNote->getMnoteOffset();
+    auto ifd = makerNote->getIfdInEntry(ORF_TAG_CAMERA_SETTINGS);
+    if (ifd) {
+        LOGDBG1("CameraSettings %lu\n", (LSIZE)ifd->entries().size());
+        uint32_t is_valid = ifd->getValue<uint32_t>(ORF_TAG_CS_PREVIEW_IMAGE_VALID).value_or(0);
+        if (is_valid) {
+            uint32_t start =
+                ifd->getValue<uint32_t>(ORF_TAG_CS_PREVIEW_IMAGE_START).value_or(0);
+            if (start) {
+                start += makerNote->getMnoteOffset();
+            }
+            uint32_t len = ifd->getValue<uint32_t>(ORF_TAG_CS_PREVIEW_IMAGE_LENGTH).value_or(0);
+            LOGDBG1("is_valid %u, start %u len %u\n", is_valid, start, len);
 
-        LOGDBG1("fetching JPEG\n");
-        IO::Stream::Ptr s(
-            std::make_shared<IO::StreamClone>(m_io, val_offset));
-        std::unique_ptr<JfifContainer> jfif(new JfifContainer(s, 0));
-
-        uint32_t x, y;
-        x = y = 0;
-        jfif->getDimensions(x, y);
-        LOGDBG1("JPEG dimensions x=%d y=%d\n", x, y);
-
-        uint32_t dim = std::max(x, y);
-        if (dim) {
-            _addThumbnail(dim, ThumbDesc(x, y, OR_DATA_TYPE_JPEG,
-                                         val_offset, e->count()));
-            list.push_back(dim);
-            err = OR_ERROR_NONE;
+            // if either value is zero we consider it is wrong.
+            if (start != 0 && len != 0) {
+                err = _addThumbnailFromStream(start, len, list);
+            } else {
+                err = OR_ERROR_NOT_FOUND;
+            }
         }
     }
 
@@ -352,21 +393,21 @@ IfdDir::Ref  OrfFile::_locateMainIfd()
         uint32_t y = data.height();
         uint32_t compression = 0;
         if(data.size() < x * y * 2) {
-            compression = ORF_COMPRESSION;
-            data.setCompression(ORF_COMPRESSION);
+            compression = IFD::COMPRESS_OLYMPUS;
+            data.setCompression(IFD::COMPRESS_OLYMPUS);
             data.setDataType(OR_DATA_TYPE_COMPRESSED_RAW);
         }
         else {
             compression = data.compression();
         }
         switch(compression) {
-        case ORF_COMPRESSION:
+        case IFD::COMPRESS_OLYMPUS:
             if((options & OR_OPTIONS_DONT_DECOMPRESS) == 0) {
                 OlympusDecompressor decomp((const uint8_t*)data.data(),
                                            data.size(), m_container, x, y);
                 RawDataPtr dData = decomp.decompress();
                 if (dData) {
-                    dData->setCfaPatternType(data.cfaPattern()->patternType());
+                    dData->setCfaPatternType(data.mosaicInfo()->patternType());
                     data.swap(*dData);
                     data.setDataType(OR_DATA_TYPE_RAW);
                     data.setDimensions(x, y);
@@ -380,14 +421,6 @@ IfdDir::Ref  OrfFile::_locateMainIfd()
     return err;
 }
 
-uint32_t OrfFile::_translateCompressionType(IFD::TiffCompress tiffCompression)
-{
-    if(tiffCompression == IFD::COMPRESS_CUSTOM) {
-        return ORF_COMPRESSION;
-    }
-    return (uint32_t)tiffCompression;
-}
-
 }
 }
 /*
