@@ -2,7 +2,7 @@
 /*
  * libopenraw - raffile.cpp
  *
- * Copyright (C) 2011-2018 Hubert Figuière
+ * Copyright (C) 2011-2020 Hubert Figuière
  *
  * This library is free software: you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -37,11 +37,13 @@
 #include "ifd.hpp"
 #include "ifddir.hpp"
 #include "ifdentry.hpp"
+#include "ifdfilecontainer.hpp"
 #include "rawfile_private.hpp"
 #include "raffile.hpp"
 #include "rafcontainer.hpp"
 #include "rafmetacontainer.hpp"
 #include "jfifcontainer.hpp"
+#include "makernotedir.hpp"
 #include "unpack.hpp"
 #include "trace.hpp"
 #include "io/streamclone.hpp"
@@ -55,6 +57,10 @@ namespace Internals {
 
 /* taken from dcraw, by default */
 static const BuiltinColourMatrix s_matrices[] = {
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_F550EXR),
+      0,
+      0,
+      { 1369, -5358, -1474, -3369, 11600, 1998, -132, 1554, 4395 } },
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_F700),
       0,
       0,
@@ -79,6 +85,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 12300, -5110, -1304, -9117, 17143, 1998, -1947, 2448, 8100 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S5000),
+      0,
+      0,
+      { 8754, -2732, -1019, -7204, 15069, 2276, -1702, 2334, 6982 } },
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S5600),
       0,
       0,
@@ -95,6 +105,11 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0xf68,
       { 12440, -3954, -1183, -1123, 9674, 1708, -83, 1614, 4086 } },
+    // HS33EXR is an alias of this.
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_HS30EXR),
+      0,
+      0,
+      { 1369, -5358, -1474, -3369, 11600, 1998, -132, 1554, 4395 } },
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100),
       0,
       0,
@@ -113,6 +128,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 11434, -4948, -1210, -3746, 12042, 1903, -666, 1479, 5235 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100V),
+      0,
+      0,
+      { 13426, -6334, -1177, -4244, 12136, 2371, -580, 1303, 5980 } },
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X10),
       0,
       0,
@@ -139,6 +158,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 11434, -4948, -1210, -3746, 12042, 1903, -666, 1479, 5235 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XPRO3),
+      0,
+      0,
+      { 13426, -6334, -1177, -4244, 12136, 2371, -580, 1303, 5980 } },
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XA1),
       0,
       0,
@@ -157,6 +180,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 11673, -476, -1041, -3988, 12058, 2166, -771, 1417, 5569 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XA7),
+      0,
+      0,
+      { 15055, -7391, -1274, -4062, 12071, 2238, -610, 1217, 6147 } },
     // From DNG Converter 10.3
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XQ1),
       0,
@@ -201,6 +228,10 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 8458, -2451, -855, -4597, 12447, 2407, -1475, 2482, 6526 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT100),
+      0,
+      0,
+      { 11673, -476, -1041, -3988, 12058, 2166, -771, 1417, 5569 } },
     // From DNG Converter 10.3
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT2),
       0,
@@ -211,35 +242,84 @@ static const BuiltinColourMatrix s_matrices[] = {
       0,
       0,
       { 11434, -4948, -1210, -3746, 12042, 1903, -666, 1479, 5235 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT200),
+      0,
+      0,
+      { 15055, -7391, -1274, -4062, 12071, 2238, -610, 1217, 6147 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT3),
+      0,
+      0,
+      { 16393, -7740, -1436, -4238, 12131, 2371, -633, 1424, 6553 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT30),
+      0,
+      0,
+      { 13426, -6334, -1177, -4244, 12136, 2371, -580, 1303, 5980 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT4),
+      0,
+      0,
+      { 13426, -6334, -1177, -4244, 12136, 2371, -580, 1303, 5980 } },
     // From DNG Converter 7.1-rc
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XS1),
       0,
       0,
       { 13509, -6199, -1254, -4430, 12733, 1865, -331, 1441, 5022 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XS10),
+      0,
+      0,
+      { 13426, -6334, -1177, -4244, 12136, 2371, -580, 1303, 5980 } },
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XF1),
       0,
       0,
       { 13509, -6199, -1254, -4430, 12733, 1865, -331, 1441, 5022 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XF10),
+      0,
+      0,
+      { 11673, -476, -1041, -3988, 12058, 2166, -771, 1417, 5569 } },
     { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S200EXR),
       512,
       0x3fff,
       { 11401, -4498, -1312, -5088, 12751, 2613, -838, 1568, 5941 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S100FS),
+      512,
+      0x3fff,
+      { 11521, -4355, -1065, -6524, 13768, 3059, -1466, 1984, 6045 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_GFX50S),
+      0,
+      0,
+      { 11756, -4754, -874, -3056, 11045, 2305, -381, 1457, 6006 } },
+    // For now we assume it is the same sensor as the GFX50S
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_GFX50R),
+      0,
+      0,
+      { 11756, -4754, -874, -3056, 11045, 2305, -381, 1457, 6006 } },
+    { OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_GFX100),
+      0,
+      0,
+      { 16212, -8423, -1583, -4336, 12583, 1937, -195, 726, 6199 } },
 
     { 0, 0, 0, { 0, 0, 0, 0, 0, 0, 0, 0, 0 } }
 };
 
 const RawFile::camera_ids_t RafFile::s_def[] = {
     { "GFX 50S", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_GFX50S) },
+    { "GFX 50R", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_GFX50R) },
+    { "GFX 100", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_GFX100) },
+    { "FinePix F550EXR", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_F550EXR) },
     { "FinePix F700  ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_F700) },
     { "FinePix F810   ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_F810) },
     { "FinePix E900   ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_E900) },
     { "FinePixS2Pro", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S2PRO) },
     { "FinePix S3Pro  ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S3PRO) },
     { "FinePix S5Pro  ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S5PRO) },
+    { "FinePix S5000 ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S5000) },
     { "FinePix S5600  ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S5600) },
     { "FinePix S9500  ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S9500) },
     { "FinePix S6500fd", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S6500FD) },
     { "FinePix HS10 HS11", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_HS10) },
+    { "FinePix HS30EXR", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_HS30EXR) },
+    { "FinePix HS33EXR", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_HS33EXR) },
+    { "FinePix S100FS ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S100FS) },
+    { "FinePix S200EXR", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S200EXR) },
     { "FinePix X100", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100) },
     { "X10", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X10) },
     { "X20", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X20) },
@@ -247,26 +327,35 @@ const RawFile::camera_ids_t RafFile::s_def[] = {
     { "X70", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X70) },
     { "X-Pro1", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XPRO1) },
     { "X-Pro2", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XPRO2) },
+    { "X-Pro3", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XPRO3) },
     { "X-S1", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XS1) },
-    { "FinePix S200EXR", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_S200EXR) },
+    { "X-S10", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XS10) },
     { "X-A1", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XA1) },
     { "X-A2", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XA2) },
     { "X-A3", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XA3) },
     { "X-A5", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XA5) },
+    { "X-A7", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XA7) },
     { "XQ1", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XQ1) },
     { "XQ2", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XQ2) },
-    { "X-E1  ", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XE1) },
+    { "X-E1", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XE1) },
     { "X-E2", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XE2) },
     { "X-E2S", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XE2S) },
     { "X-M1", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XM1) },
     { "X-T1", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT1) },
     { "X-T10", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT10) },
+    { "X-T100", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT100) },
     { "X-T2", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT2) },
     { "X-T20", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT20) },
+    { "X-T200", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT200) },
+    { "X-T3", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT3) },
+    { "X-T30", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT30) },
+    { "X-T4", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT4) },
     { "XF1", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XF1) },
+    { "XF10", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XF10) },
     { "X100S", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100S) },
     { "X100T", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100T) },
     { "X100F", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100F) },
+    { "X100V", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100V) },
     { "X-E3", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XE3) },
     { "X-H1", OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XH1) },
 
@@ -290,6 +379,18 @@ RafFile::~RafFile()
     delete m_container;
 }
 
+IfdDir::Ref RafFile::_locateMainIfd()
+{
+    if (!m_mainIfd) {
+        JfifContainer *jpegPreview = m_container->getJpegPreview();
+        if (!jpegPreview) {
+            return IfdDir::Ref();
+        }
+        m_mainIfd = jpegPreview->getIfdDirAt(0);
+    }
+    return m_mainIfd;
+}
+
 ::or_error RafFile::_enumThumbnailSizes(std::vector<uint32_t> &list)
 {
     or_error ret = OR_ERROR_NOT_FOUND;
@@ -314,6 +415,8 @@ RafFile::~RafFile()
         return ret;
     }
 
+    // XXX check why this as it appear that if true there won't be
+    // and thumbnail.
     auto result = dir->getIntegerValue(IFD::EXIF_TAG_IMAGE_WIDTH);
     if (result) {
         x = result.value();
@@ -327,19 +430,16 @@ RafFile::~RafFile()
         if (result.empty()) {
             return ret;
         }
-        uint32_t jpeg_offset = result.value();
+        uint32_t jpeg_offset = result.value() + jpegPreview->exifOffset();
 
-        jpeg_offset += 12; // XXX magic number. eh?
-                           // I need to re-read the Exif spec.
         result = dir->getValue<uint32_t>(IFD::EXIF_TAG_JPEG_INTERCHANGE_FORMAT_LENGTH);
         if (result.empty()) {
             return ret;
         }
         uint32_t jpeg_size = result.value();
 
-        IO::Stream::Ptr s(std::make_shared<IO::StreamClone>(jpegPreview->file(),
-                                                            jpeg_offset));
-        std::unique_ptr<JfifContainer> thumb(new JfifContainer(s, 0));
+        auto s = std::make_shared<IO::StreamClone>(jpegPreview->file(), jpeg_offset);
+        auto thumb = std::make_unique<JfifContainer>(s, 0);
 
         if (thumb->getDimensions(x, y)) {
             uint32_t size = std::max(x, y);
@@ -367,6 +467,7 @@ RafFile::isXTrans(RawFile::TypeId type_) const
     switch (type_) {
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XPRO1):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XPRO2):
+    case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XPRO3):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XE1):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XE2):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XE2S):
@@ -379,12 +480,16 @@ RafFile::isXTrans(RawFile::TypeId type_) const
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT10):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT2):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT20):
+    case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT3):
+    case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XT4):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100S):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100T):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100F):
+    case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X100V):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X20):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X30):
     case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_X70):
+    case OR_MAKE_FUJIFILM_TYPEID(OR_TYPEID_FUJIFILM_XS10):
         return true;
 
     default:
@@ -407,12 +512,12 @@ RafFile::isXTrans(RawFile::TypeId type_) const
         // use this tag if the other is missing
         value = meta->getValue(RAF_TAG_IMG_HEIGHT_WIDTH);
     }
-    uint32_t dims = value->get().getInteger(0);
+    uint32_t dims = value->get().getUInteger(0);
     uint16_t h = (dims & 0xFFFF0000) >> 16;
     uint16_t w = (dims & 0x0000FFFF);
 
     value = meta->getValue(RAF_TAG_RAW_INFO);
-    uint32_t rawProps = value->get().getInteger(0);
+    uint32_t rawProps = value->get().getUInteger(0);
     // TODO re-enable if needed.
     // uint8_t layout = (rawProps & 0xFF000000) >> 24 >> 7; // MSBit in byte.
     uint8_t compressed = ((rawProps & 0xFF0000) >> 16) & 8; // 8 == compressed
@@ -422,7 +527,7 @@ RafFile::isXTrans(RawFile::TypeId type_) const
     data.setDataType(OR_DATA_TYPE_RAW);
     data.setDimensions(w, h);
     if (isXTrans(typeId())) {
-        data.setCfaPattern(XTransPattern::xtransPattern());
+        data.setMosaicInfo(XTransPattern::xtransPattern());
     } else {
         data.setCfaPatternType(OR_CFA_PATTERN_GBRG);
     }
@@ -437,7 +542,7 @@ RafFile::isXTrans(RawFile::TypeId type_) const
     uint32_t datalen = (is_compressed ? byte_size : finaldatalen);
     void *buf = data.allocData(finaldatalen);
 
-    LOGDBG2("byte_size = %lu finaldatalen = %u compressed = %u", byte_size,
+    LOGDBG2("byte_size = %lu finaldatalen = %u compressed = %u", (LSIZE)byte_size,
             finaldatalen, compressed);
 
     ret = OR_ERROR_NONE;
@@ -446,7 +551,7 @@ RafFile::isXTrans(RawFile::TypeId type_) const
         Unpack unpack(w, IFD::COMPRESS_NONE);
         size_t blocksize = unpack.block_size();
         std::unique_ptr<uint8_t[]> block(new uint8_t[blocksize]);
-        uint8_t *outdata = (uint8_t *)data.data();
+        uint16_t *outdata = (uint16_t *)data.data();
         size_t outsize = finaldatalen;
         size_t got;
         do {
@@ -458,7 +563,7 @@ RafFile::isXTrans(RawFile::TypeId type_) const
                 size_t out;
                 or_error err = unpack.unpack_be12to16(outdata, outsize,
                                                       block.get(), got, out);
-                outdata += out;
+                outdata += out / 2;
                 outsize -= out;
                 if (err != OR_ERROR_NONE) {
                     LOGDBG2("error is %d\n", static_cast<int>(err));
@@ -488,7 +593,7 @@ MetaValue *RafFile::_getMetaValue(int32_t meta_index)
         IfdDir::Ref dir = jpegPreview->mainIfd();
         IfdEntry::Ref e = dir->getEntry(META_NS_MASKOUT(meta_index));
         if (e) {
-            return e->make_meta_value();
+            return dir->makeMetaValue(*e);
         }
     }
 
@@ -497,7 +602,7 @@ MetaValue *RafFile::_getMetaValue(int32_t meta_index)
 
 void RafFile::_identifyId()
 {
-    _setTypeId(_typeIdFromModel("FujiFilm", m_container->getModel()));
+    _setTypeId(_typeIdFromModel("FUJIFILM", m_container->getModel()));
 }
 }
 }
