@@ -2,7 +2,7 @@
 /*
  * libopenraw - rawcontainer.h
  *
- * Copyright (C) 2006-2017 Hubert Figuière
+ * Copyright (C) 2006-2020 Hubert Figuière
  *
  * This library is free software: you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -19,8 +19,7 @@
  * <http://www.gnu.org/licenses/>.
  */
 
-#ifndef OR_INTERNALS_RAWCONTAINER_H_
-#define OR_INTERNALS_RAWCONTAINER_H_
+#pragma once
 
 #include <stddef.h>
 #include <stdint.h>
@@ -56,27 +55,32 @@ public:
     /** destructor */
     virtual ~RawContainer();
 
-    const IO::Stream::Ptr &file() { return m_file; }
+    const IO::Stream::Ptr &file() const { return m_file; }
     EndianType endian() const { return m_endian; }
     off_t offset() const { return m_offset; }
+    // This only make sense for IFD.
+    virtual int exifOffsetCorrection() const
+        {
+            return 0;
+        }
 
     bool skip(off_t offset);
-    Option<int8_t> readInt8(const IO::Stream::Ptr &f);
-    Option<uint8_t> readUInt8(const IO::Stream::Ptr &f);
+    Option<int8_t> readInt8(const IO::Stream::Ptr& f) const;
+    Option<uint8_t> readUInt8(const IO::Stream::Ptr& f) const;
     /** Read an int16 following the m_endian set */
-    Option<int16_t> readInt16(const IO::Stream::Ptr &f);
+    Option<int16_t> readInt16(const IO::Stream::Ptr& f, EndianType endian) const;
     /** Read an int32 following the m_endian set */
-    Option<int32_t> readInt32(const IO::Stream::Ptr &f);
+    Option<int32_t> readInt32(const IO::Stream::Ptr& f, EndianType endian) const;
     /** Read an uint16 following the m_endian set */
-    Option<uint16_t> readUInt16(const IO::Stream::Ptr &f);
+    Option<uint16_t> readUInt16(const IO::Stream::Ptr& f, EndianType endian) const;
     /** Read an array of uint16 following the m_endian set.
      * @param v the vector to fill. Will be resized if too small.
      * @param count the number of elements to read
      * @return the number of element read. `count` if success.
      */
-    size_t readUInt16Array(const IO::Stream::Ptr &f, std::vector<uint16_t> &v, size_t count);
+    size_t readUInt16Array(const IO::Stream::Ptr& f, std::vector<uint16_t>& v, size_t count) const;
     /** Read an uint32 following the m_endian set */
-    Option<uint32_t> readUInt32(const IO::Stream::Ptr &f);
+    Option<uint32_t> readUInt32(const IO::Stream::Ptr& f, EndianType endian) const;
     /**
      * Fetch the data chunk from the file
      * @param buf the buffer to load into
@@ -84,7 +88,7 @@ public:
      * @param buf_size the size of the data to fetch
      * @return the size retrieved, <= buf_size likely equal
      */
-    size_t fetchData(void *buf, off_t offset, size_t buf_size);
+    size_t fetchData(void* buf, off_t offset, size_t buf_size) const;
 
     /**
      * Return the effective size of the container.
@@ -92,6 +96,7 @@ public:
     off_t size() const;
 protected:
     RawContainer(const RawContainer &) = delete;
+    RawContainer(RawContainer&&) = default;
     RawContainer &operator=(const RawContainer &) = delete;
 
     void setEndian(EndianType _endian) { m_endian = _endian; }
@@ -104,5 +109,3 @@ protected:
 };
 }
 }
-
-#endif
